﻿using System;
using System.Collections.Generic;
using System.Collections.ObjectModel;
using System.Linq;
using System.Threading.Tasks;
using ThinkGeo.MapSuite.Core;
using ThinkGeo.MapSuite.Routing;

namespace SmartBruteForceRoutingForTsp
{
    public class BruteForceRoutingEngine
    {
        private FeatureSource featureSource;
        private RoutingSource routingSource;
        private RoutingEngine routingEngine;
        private Dictionary<Tuple<PointShape, PointShape>, RoutingResult> distanceMatrix;
        private Collection<Collection<Tuple<RoutingResult, PointShape>>> bestRoutes;

        public BruteForceRoutingEngine()
            : this(new ShapeFileFeatureSource(), new RtgRoutingSource())
        { }

        public BruteForceRoutingEngine(FeatureSource featureSource, RoutingSource routingSource)
        {
            this.featureSource = featureSource;
            this.routingSource = routingSource;
            routingEngine = new RoutingEngine(routingSource, new AStarRoutingAlgorithm(), featureSource) { SearchRadiusInMeters = 50 };
            distanceMatrix = new Dictionary<Tuple<PointShape, PointShape>, RoutingResult>();
            bestRoutes = new Collection<Collection<Tuple<RoutingResult, PointShape>>>();
        }

        public RoutingResult GetRoute(PointShape startPoint, PointShape endPoint, IEnumerable<PointShape> stops)
        {
            return GetRouteCore(startPoint, endPoint, stops, 5, GetOptimalNumberOfSplits(stops.Count()));
        }

        public RoutingResult GetRoute(PointShape startPoint, PointShape endPoint, IEnumerable<PointShape> stops, int maximumConcurrentThreads)
        {
            return GetRouteCore(startPoint, endPoint, stops, maximumConcurrentThreads, GetOptimalNumberOfSplits(stops.Count()));
        }

        private RoutingResult GetRouteCore(PointShape startPoint, PointShape endPoint, IEnumerable<PointShape> stops, int maximumConcurrentThreads, int numberOfSplits)
        {
            bestRoutes = new Collection<Collection<Tuple<RoutingResult, PointShape>>>();

            Collection<PointShape> pointsToVisit = new Collection<PointShape>();
            foreach (PointShape pointToVisit in stops)
            {
                pointsToVisit.Add(pointToVisit);
            }

            Collection<PointShape> allPoints = pointsToVisit.DeepClone();
            allPoints.Add(startPoint);
            allPoints.Add(endPoint);

            InitializeDistanceMatrix(allPoints);

            Collection<Collection<PointShape>> intermediatePointSets = GetIntermediatePoints(startPoint, endPoint, pointsToVisit, numberOfSplits);

            Parallel.ForEach(intermediatePointSets, new ParallelOptions { MaxDegreeOfParallelism = maximumConcurrentThreads }, intermediatePointSet =>
            {
                Collection<PointShape> pointsToVisitCopy = pointsToVisit.DeepClone();
                foreach (PointShape point in intermediatePointSet)
                {
                    pointsToVisitCopy.Remove(point);
                }

                GetBestRoute(startPoint, intermediatePointSet, pointsToVisitCopy);
            });

            double minimumDistance = double.MaxValue;
            Collection<Tuple<RoutingResult, PointShape>> bestRoute = new Collection<Tuple<RoutingResult, PointShape>>();

            foreach (Collection<Tuple<RoutingResult, PointShape>> route in bestRoutes)
            {
                double localMinimumDistance = distanceMatrix[Tuple.Create(startPoint, route[0].Item2)].Distance;

                for (int i = 0; i < route.Count - 1; i++)
                {
                    localMinimumDistance += distanceMatrix[Tuple.Create(route[i].Item2, route[i + 1].Item2)].Distance;
                }

                if (localMinimumDistance < minimumDistance)
                {
                    minimumDistance = localMinimumDistance;
                    bestRoute = route;
                }
            }

            RoutingResult routingResult = new RoutingResult();
            routingResult.Distance = minimumDistance;
            routingResult.Weight = minimumDistance;

            for (int i = 0; i < bestRoute.Count; i++)
            {
                foreach (var line in bestRoute[i].Item1.RouteLines.Lines)
                {
                    routingResult.RouteLines.Lines.Add(line);
                }

                foreach (var feature in bestRoute[i].Item1.Features)
                {
                    routingResult.Features.Add(feature);
                }

                foreach (var segment in bestRoute[i].Item1.RouteSegments)
                {
                    routingResult.RouteSegments.Add(segment);
                }

                foreach (var vertex in bestRoute[i].Item1.Route.Vertices)
                {
                    routingResult.Route.Vertices.Add(vertex);
                }

                routingResult.OrderedStops.Add(bestRoute[i].Item2);
            }
            routingResult.OrderedStops.RemoveAt(routingResult.OrderedStops.Count - 1);

            return routingResult;
        }

        private Collection<Collection<PointShape>> GetIntermediatePoints(PointShape startPoint, PointShape endPoint, Collection<PointShape> pointsToVisit, int numberOfSplits)
        {
            Collection<PointShape> intermediatePoints = new Collection<PointShape>();
            intermediatePoints.Add(startPoint);
            intermediatePoints.Add(endPoint);

            Collection<Collection<PointShape>> intermediatePointSets = new Collection<Collection<PointShape>>();
            int possibleIntermediatePointSetSize = (int)(((double)1 / (double)numberOfSplits) * pointsToVisit.Count);
            Collection<PointShape> possibleIntermediatePoints = new Collection<PointShape>();
            pointsToVisit.OrderByDescending(point => (distanceMatrix[Tuple.Create(point, startPoint)].Distance / distanceMatrix[Tuple.Create(endPoint, point)].Distance)).Skip(pointsToVisit.Count / (numberOfSplits * 2)).ToList().ForEach(e => possibleIntermediatePoints.Add(e));

            return GetIntermediatePointsRecursively(startPoint, intermediatePoints, possibleIntermediatePoints, intermediatePointSets, possibleIntermediatePointSetSize, numberOfSplits - 1);
        }

        private Collection<Collection<PointShape>> GetIntermediatePointsRecursively(PointShape startPoint, Collection<PointShape> intermediatePoints, Collection<PointShape> possibleIntermediatePoints, Collection<Collection<PointShape>> intermediatePointSets, int intermediatePointSetSize, int numberOfSplitsLeft)
        {
            if (numberOfSplitsLeft <= 0)
            {
                intermediatePointSets.Add(intermediatePoints);
            }
            else
            {
                foreach (PointShape intermediatePoint in possibleIntermediatePoints.Skip(intermediatePointSetSize * (numberOfSplitsLeft - 1)).Take(intermediatePointSetSize))
                {
                    Collection<PointShape> possibleIntermediatePointsCopy = possibleIntermediatePoints.DeepClone();
                    Collection<PointShape> intermediatePointsCopy = intermediatePoints.DeepClone();

                    possibleIntermediatePointsCopy.Remove(intermediatePoint);
                    intermediatePointsCopy.Insert(intermediatePoints.Count - 1, intermediatePoint);

                    intermediatePointSets = GetIntermediatePointsRecursively(startPoint, intermediatePointsCopy, possibleIntermediatePointsCopy, intermediatePointSets, intermediatePointSetSize, numberOfSplitsLeft - 1);
                }
            }

            return intermediatePointSets;
        }

        private void GetBestRoute(PointShape startPoint, Collection<PointShape> intermediatePoints, Collection<PointShape> pointsToVisit)
        {
            double minimumDistance = double.MaxValue;
            Collection<Tuple<RoutingResult, PointShape>> bestRoute = new Collection<Tuple<RoutingResult, PointShape>>();

            Collection<PointShape> points = pointsToVisit.DeepClone();

            for (int i = 0; i < intermediatePoints.Count - 1; i++)
            {
                Collection<PointShape> nextSetOfPoints = new Collection<PointShape>();
                if (i == intermediatePoints.Count - 2)
                {
                    nextSetOfPoints = points.DeepClone();
                }
                else
                {
                    points.OrderByDescending(point => (distanceMatrix[Tuple.Create(point, intermediatePoints.Last())].Distance / distanceMatrix[Tuple.Create(startPoint, point)].Distance)).Take(points.Count / (intermediatePoints.Count - i - 1)).ToList().ForEach(e => { nextSetOfPoints.Add(e); points.Remove(e); });
                }

                Tuple<Collection<Tuple<RoutingResult, PointShape>>, double> result = FindBestRoute(intermediatePoints[i], intermediatePoints[i + 1], nextSetOfPoints);

                foreach (Tuple<RoutingResult, PointShape> routeResult in result.Item1)
                {
                    bestRoute.Add(routeResult);
                }
                minimumDistance += result.Item2;
            }

            if (!(bestRoute == null))
            {
                bestRoutes.Add(bestRoute);
            }
        }

        private Tuple<Collection<Tuple<RoutingResult, PointShape>>, double> FindBestRoute(PointShape startPoint, PointShape endPoint, Collection<PointShape> pointsToTraverse)
        {
            return FindBestRouteRecursively(startPoint, endPoint, pointsToTraverse, new Collection<Tuple<RoutingResult, PointShape>>(), 0, new Collection<Tuple<RoutingResult, PointShape>>(), double.MaxValue);
        }

        private Tuple<Collection<Tuple<RoutingResult, PointShape>>, double> FindBestRouteRecursively(
            PointShape currentPoint,
            PointShape endPoint,
            Collection<PointShape> pointsToTraverse,
            Collection<Tuple<RoutingResult, PointShape>> currentRoute,
            double currentDistance,
            Collection<Tuple<RoutingResult, PointShape>> localBestRoute,
            double localMinDistance)
        {
            Collection<PointShape> closestPoints = new Collection<PointShape>();

            if (pointsToTraverse.Count == 0)
            {
                currentDistance += distanceMatrix[Tuple.Create(currentPoint, endPoint)].Distance;
                if (currentDistance >= localMinDistance)
                {
                    return Tuple.Create(localBestRoute, localMinDistance);
                }
                else if (currentDistance < localMinDistance)
                {
                    currentRoute.Add(Tuple.Create(distanceMatrix[Tuple.Create(currentPoint, endPoint)], endPoint));

                    return Tuple.Create(currentRoute, currentDistance);
                }
            }
            else if (pointsToTraverse.Count < 3)
            {
                closestPoints = pointsToTraverse.DeepClone();
            }
            else
            {
                closestPoints = new Collection<PointShape>();
                pointsToTraverse.OrderByDescending(i => distanceMatrix[Tuple.Create(currentPoint, i)].Distance).Reverse().Take(3).ToList().ForEach(e => closestPoints.Add(e));
            }

            foreach (PointShape point in closestPoints)
            {
                double currentDistanceCopy = currentDistance;
                currentDistanceCopy += distanceMatrix[Tuple.Create(currentPoint, point)].Distance;
                if (currentDistanceCopy > localMinDistance)
                {
                    return Tuple.Create(localBestRoute, localMinDistance);
                }

                Collection<PointShape> pointsToTraverseCopy = pointsToTraverse.DeepClone();
                pointsToTraverseCopy.Remove(point);

                Collection<Tuple<RoutingResult, PointShape>> currentRouteCopy = currentRoute.DeepClone();
                currentRouteCopy.Add(Tuple.Create(distanceMatrix[Tuple.Create(currentPoint, point)], point));

                Tuple<Collection<Tuple<RoutingResult, PointShape>>, double> results = FindBestRouteRecursively(point, endPoint, pointsToTraverseCopy, currentRouteCopy, currentDistanceCopy, localBestRoute, localMinDistance);
                localBestRoute = results.Item1;
                localMinDistance = results.Item2;
            }

            return Tuple.Create(localBestRoute, localMinDistance);
        }

        private void InitializeDistanceMatrix(IEnumerable<PointShape> stops)
        {
            distanceMatrix = new Dictionary<Tuple<PointShape, PointShape>, RoutingResult>();

            foreach (PointShape fromPoint in stops)
            {
                foreach (PointShape toPoint in stops)
                {
                    if (!fromPoint.Equals(toPoint))
                    {
                        if (distanceMatrix.ContainsKey(Tuple.Create(toPoint, fromPoint)))
                        {
                            distanceMatrix.Add(Tuple.Create(fromPoint, toPoint), distanceMatrix[Tuple.Create(toPoint, fromPoint)]);
                        }
                        else
                        {
                            RoutingResult tempResult = routingEngine.GetRoute(fromPoint, toPoint);
                            distanceMatrix.Add(Tuple.Create(fromPoint, toPoint), tempResult);
                        }
                    }
                }
            }
        }

        private int GetOptimalNumberOfSplits(int stops)
        {
            int numberOfSplits = 1;
            long total = stops * (long)Math.Pow(3, stops);
            long nextSplitTotal = stops * (long)Math.Pow(3, stops);

            while (total >= nextSplitTotal)
            {
                numberOfSplits++;
                total = nextSplitTotal;

                nextSplitTotal = stops * (long)Math.Pow(3, stops / numberOfSplits);
                for (int i = 0; i < numberOfSplits; i++)
                {
                    nextSplitTotal = nextSplitTotal * (stops - i + 2);
                }
            }
            return numberOfSplits - 1;
        }
    }

    static class Extensions
    {
        public static Collection<T> DeepClone<T>(this Collection<T> collectionToCopy)
        {
            Collection<T> copy = new Collection<T>();
            foreach (T item in collectionToCopy)
            {
                copy.Add(item);
            }

            return copy;
        }
    }
}
